using System;
using System.Threading;

class CustomReaderWriter
{
	const int READERS_COUNT = 5;
	const int WRITERS_COUNT = 2;

	private int mReadersCount = 0;
	private bool mIsWriting = false;
	private string mSharedResource = "00";

	private Random mRand = new Random();

	public void Read()
	{
		while (true)
		{
			Thread.Sleep(mRand.Next(5000));
			
			// Wait for the reading to become allowed
			// and register current reader
			lock (this)
			{
				while (mIsWriting)
				{
					Monitor.Wait(this);
				}
				mReadersCount++;
			}

			Console.WriteLine( "READING...data: {0} readers: {1}, writers: {2}",
				mSharedResource, mReadersCount, mIsWriting ? 1 : 0 );

			Thread.Sleep(mRand.Next(5000));

			// Unregister current reader
			lock (this)
			{
				mReadersCount--;
				if (mReadersCount == 0)
				{
					Monitor.Pulse(this);
				}
			}

			Thread.Sleep(mRand.Next(5000));
		}
	}

	public void Write()
	{
		while(true)
		{
			// Wait for the writing to become allowed
			// and register current writer
			lock (this)
			{
				while (mReadersCount != 0)
				{
					Monitor.Wait( this );
				}
				mIsWriting = true;
			}

			Console.WriteLine("WRITTING...data: {0} readers: {1}, writers: {2}",
				mSharedResource, mReadersCount, mIsWriting ? 1 : 0);

			int newValue = mRand.Next(89) + 10;
			mSharedResource = newValue.ToString();

			Thread.Sleep(mRand.Next(5000));

			// Unregister current writer
			lock (this)
			{
				mIsWriting = false;
				Monitor.PulseAll(this);
			}

			Thread.Sleep(mRand.Next(5000));
		}
	}

	static void Main()
	{
		CustomReaderWriter customReaderWriter = new CustomReaderWriter();

		for( int i=0; i<READERS_COUNT; i++ )
		{
			Thread thread = new Thread( 
				new ThreadStart(customReaderWriter.Read) );
			thread.Start();
		}

		for( int i=0; i<WRITERS_COUNT; i++ )
		{
			Thread thread = new Thread( 
				new ThreadStart(customReaderWriter.Write) );
			thread.Start();
		}
	}
}